"""
Authors: Peter Mawhorter
Consulted:
Date: 2022-2-5
Purpose: Tests for treeTops task.
"""

import optimism as opt
import treeTops

import turtleBeads
import turtle

import math # for trig


# Reset turtle window, especially in case importing treeTops drew
# something or moved the turtle.
turtleBeads.setupTurtle()


#---------#
# Helpers #
#---------#

def drawHeight(x, y):
    """
    Draws a short horizontal dotted line centered on the given x
    coordinate and at the given y coordinate, to mark a height.

    Changes the position and heading of the turtle in the process, as
    well as the pen size and color.
    """
    turtle.pensize(2)
    turtle.pencolor('blue')
    turtleBeads.teleport(x - 20, y)
    turtle.setheading(0)
    # Description
    turtleBeads.describeAs(f"Reported top height is: {round(y, 3)}")
    # Dashed line
    dash = 40 / 7
    turtle.fd(dash)
    turtleBeads.leap(dash)
    turtle.fd(dash)
    turtleBeads.leap(dash)
    turtle.fd(dash)
    turtleBeads.leap(dash)
    turtle.fd(dash)
    turtleBeads.endDescription()


#-------#
# Tests #
#-------#

# This factor is multiplied by a length to get the height of something
# of that length at a 20-degree angle off of the vertical.
angleFactor = math.cos((20 / 180) * math.pi)

# Set up for testing: we want to draw fast (we're not testing what's
# drawn, only what's printed)
turtleBeads.noTrace()


#----------------#
# Tests for twig #
#----------------#

# Head west and face upwards
turtleBeads.teleport(-200, 0)
turtle.setheading(90)
testTwig = opt.testFunctionMaybe(treeTops, "twig")
case = testTwig.case(3, 30, 'Sienna', 'SpringGreen2')
case.checkReturnValue(30 + 15 * angleFactor)

# Check the printed output to find the lines drawn, in any order
case.checkPrintedFragment('A 3-pensize Sienna vertical line')
# Don't know order for these
case.checkPrintedFragment('(-200, 0)') # twig base
case.checkPrintedFragment('(-200, 30)', 3) # leaf/twig joint
case.checkPrintedFragment('(-205, 44)') # leaf tip
case.checkPrintedFragment('(-195, 44)') # leaf tip
case.checkPrintedFragment('A 7.5-pensize SpringGreen2 110° line')
case.checkPrintedFragment('A 7.5-pensize SpringGreen2 70° line')

# Draw a dotted line at the reported top height
results = case.fetchResults()
if "result" in results:
    top = results["result"]
    drawHeight(-200, top)

# Nearby facing downwards
turtleBeads.teleport(-160, 20)
turtle.setheading(-90)
testTwig2 = opt.testFunctionMaybe(treeTops, "twig")
case = testTwig2.case(4, 40, 'Sienna', 'SpringGreen2')
case.checkReturnValue(20) # top is at the root

# Check the printed output to find the lines drawn, in any order
case.checkPrintedFragment('A 4-pensize Sienna vertical line')
case.checkPrintedFragment('(-160, 20)')
case.checkPrintedFragment('(-160, -20)', 3)
case.checkPrintedFragment('(-153, -39)')
case.checkPrintedFragment('(-167, -39)')
case.checkPrintedFragment('A 10-pensize SpringGreen2 110° line')
case.checkPrintedFragment('A 10-pensize SpringGreen2 70° line')

# Draw a dotted line at the reported top height
results = case.fetchResults()
if "result" in results:
    top = results["result"]
    drawHeight(-160, top)


#-----------------#
# Test for branch #
#-----------------#

# Off-center and facing 160 degrees
turtleBeads.teleport(50, 0)
turtle.setheading(160)
testBranch = opt.testFunctionMaybe(treeTops, "branch")
case = testBranch.case(5, 60, 40, 'Cornsilk2', 'DarkOliveGreen2')
case.checkReturnValue(
    60 * math.sin((20 / 180) * math.pi)
  + 40 * math.sin((60 / 180) * math.pi)
  + 20 * math.sin((80 / 180) * math.pi)
)

# One branch + 3 twigs
case.checkPrintedFragment('A 5-pensize Cornsilk2', 4)
# Six leaves
case.checkPrintedFragment('A 10-pensize DarkOliveGreen2', 6)
# The points touched
case.checkPrintedFragment('(50, 0)')
case.checkPrintedFragment('(-6, 21)', 4)
case.checkPrintedFragment('(-44, 7)', 3)
case.checkPrintedFragment('(-59, -6)')
case.checkPrintedFragment('(-64, 7)')
case.checkPrintedFragment('(-44, 34)', 3)
case.checkPrintedFragment('(-64, 34)')
case.checkPrintedFragment('(-59, 47)')
case.checkPrintedFragment('(-26, 55)', 3)
case.checkPrintedFragment('(-42, 68)')
case.checkPrintedFragment('(-30, 75)')
# The angles of each line (same regardless of line direction)
case.checkPrintedFragment('40° line', 3)
case.checkPrintedFragment('160° line', 2)
case.checkPrintedFragment('20° line', 2)
case.checkPrintedFragment('horizontal line', 2)
case.checkPrintedFragment('140° line', 2)
case.checkPrintedFragment('120° line')
case.checkPrintedFragment('100° line')

# Draw a dotted line at the reported top height
results = case.fetchResults()
if "result" in results:
    top = results["result"]
    drawHeight(-30, top)


#---------------#
# Test for tree #
#---------------#

# East and a bit down (tree should set heading)
# Offsets to avoid ending up in places where rounding error throws off
# printed digits...
turtleBeads.teleport(180.105, -40.105)
turtle.setheading(-10)
testTree = opt.testFunctionMaybe(treeTops, "tree")
case = testTree.case(
    90, # facing straight upwards
    8, # trunk diameter
    70, # trunk length
    45, # branch length
    25, # twig length
    "DarkOrange4", # bark color
    "Chartreuse3" # leaf color
)
# Check for the correct result
case.checkReturnValue(-40.105 + 70 + angleFactor * (45 + 25) + 12.5)
# Check for certain lines of output (in any order)
# The trunk
case.checkPrintedFragment('A 8-pensize DarkOrange4 vertical line')

# 24 leaves
case.checkPrintedFragment('A 6.2-pensize Chartreuse3', 24)
# 12 twigs + 4 branches
case.checkPrintedFragment('A 4-pensize DarkOrange4', 16)

# Line angles within the tree (same in either direction)
case.checkPrintedFragment('110° line', 4)
case.checkPrintedFragment('130° line', 5)
case.checkPrintedFragment('150° line', 4)
case.checkPrintedFragment('170° line', 4)
case.checkPrintedFragment('10° line', 8)
case.checkPrintedFragment('30° line', 9)
case.checkPrintedFragment('50° line', 9)
case.checkPrintedFragment('70° line', 8)
case.checkPrintedFragment('vertical line', 7)

# Points within the tree
# Note: we can't check larger phrases including from .. to .. to simplify
# these checks, because it's valid to draw the lines in the opposite
# direction, thus swapping the from/to fragments in the printed output
case.checkPrintedFragment('(180, -40)') # trunk base
case.checkPrintedFragment('(180, 30)', 5) # one trunk/branches joint

case.checkPrintedFragment('(219, 52)', 4) # four branch/twigs joints
case.checkPrintedFragment('(195, 72)', 4)
case.checkPrintedFragment('(165, 72)', 4)
case.checkPrintedFragment('(141, 52)', 4)

case.checkPrintedFragment('(117, 48)', 3) # twelve twig/leaves joints
case.checkPrintedFragment('(119, 65)', 3)
case.checkPrintedFragment('(133, 76)', 3)
case.checkPrintedFragment('(143, 85)', 3)
case.checkPrintedFragment('(156, 96)', 3)
case.checkPrintedFragment('(173, 96)', 3)
case.checkPrintedFragment('(187, 96)', 3)
case.checkPrintedFragment('(204, 96)', 3)
case.checkPrintedFragment('(217, 85)', 3)
case.checkPrintedFragment('(228, 76)', 3)
case.checkPrintedFragment('(241, 65)', 3)
case.checkPrintedFragment('(244, 48)', 3)

case.checkPrintedFragment('(104, 50)')
case.checkPrintedFragment('(106, 42)')
case.checkPrintedFragment('(107, 67)')
case.checkPrintedFragment('(111, 74)')
case.checkPrintedFragment('(125, 85)')
case.checkPrintedFragment('(131, 87)')
case.checkPrintedFragment('(135, 94)')
case.checkPrintedFragment('(148, 105)')
case.checkPrintedFragment('(156, 108)')
case.checkPrintedFragment('(173, 108)')
case.checkPrintedFragment('(179, 105)')
case.checkPrintedFragment('(181, 105)')
case.checkPrintedFragment('(187, 108)')
case.checkPrintedFragment('(204, 108)')
case.checkPrintedFragment('(212, 105)')
case.checkPrintedFragment('(225, 94)')
case.checkPrintedFragment('(228, 88)')
case.checkPrintedFragment('(229, 87)')
case.checkPrintedFragment('(236, 85)')
case.checkPrintedFragment('(249, 74)')
case.checkPrintedFragment('(253, 67)')
case.checkPrintedFragment('(255, 42)')
case.checkPrintedFragment('(256, 50)')

# Draw a dotted line at the reported top height
results = case.fetchResults()
if "result" in results:
    top = results["result"]
    drawHeight(180, top)

# Make sure everything shows up
turtle.hideturtle()
turtleBeads.showPicture()
